/*
 * Version: MPL 1.1 / GPLv3+ / LGPLv3+
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License or as specified alternatively below. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (C) 2011 Péter Rabi <prabi@caesar.elte.hu> (initial developer)
 *
 * All Rights Reserved.
 *
 * For minor contributions see the git repository.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 3 or later (the "GPLv3+"), or
 * the GNU Lesser General Public License Version 3 or later (the "LGPLv3+"),
 * in which case the provisions of the GPLv3+ or the LGPLv3+ are applicable
 * instead of those above.
 */

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <vector>

#include <dirent.h>

static const size_t MAXLEN = 255;

using namespace std;

/**
  * The goal of LibFilter is to find "unused direct dependencies" in
  * the binaries of LibreOffice with `ldd'. It also modifies the proper
  * makefiles to exclude the found unnecessary library dependecies.
  * The tool is expected to operate on the 3.5 branch of LibO under Linux
  * with default build configuration plus --enable-symbols. Processes
  * the GnuMake-ified subset of LibO. Expects a `make all' has been done.
  */
class LibFilter
{
public:
    int         run(char*);

private:
    enum        LibOrBin {LIB, BIN};

    bool        iterateOnDirs();
    void        processDirEntry(dirent*, char*);
    void        processBinary(char*);
    char*       getBinNameIfProperMakefile(dirent*);
    bool        getUnusedLibraries(char*, vector<char*>&);
    void        updateMakefile(char*, vector<char*>&);
    inline bool isLibNameProper(char*, size_t);

    char*       msLiboPath;
    char*       mpBinPaths[2];
    size_t      mnLiboPathLen;
    size_t      mnBinPathLen;
};

int main(int argc, char** argv)
{
    // Check if the path of LibO's core repo is provided
    if (argc < 2)
    {
        printf("Correct usage: libfilter /path/of/LibreOffice/core\n");
        return 1;
    }

    LibFilter aLibFilter;
    return aLibFilter.run(argv[argc - 1]);
}

/**
  * The main function of LibFilter, which does the necessary initialisation.
  * @param   sLiboPath   The path of the core repo of LibreOffice.
  * @return  Signals failure or success in the POSIX manner.
  */
int LibFilter::run(char* sLiboPath)
{
    // Put a '/' after sLiboPath if there is not one
    mnLiboPathLen = strlen(sLiboPath);
    if (sLiboPath[mnLiboPathLen - 1] != '/')
    {
        msLiboPath = new char[mnLiboPathLen + 2];
        strcpy(msLiboPath, sLiboPath);
        strcat(msLiboPath, "/");
        ++mnLiboPathLen;
    }
    else
    {
        msLiboPath = new char[mnLiboPathLen + 1];
        strcpy(msLiboPath, sLiboPath);
    }

    mnBinPathLen = mnLiboPathLen + 28;
    mpBinPaths[LIB] = new char[mnBinPathLen + 1];
    mpBinPaths[BIN] = new char[mnBinPathLen + 1];

    strcpy(mpBinPaths[LIB], msLiboPath);
    strcat(mpBinPaths[LIB], "solver/350/unxlngi6.pro/lib/");
    strcpy(mpBinPaths[BIN], msLiboPath);
    strcat(mpBinPaths[BIN], "solver/350/unxlngi6.pro/bin/");

    bool bRet = iterateOnDirs();

    delete[] msLiboPath;
    delete[] mpBinPaths[LIB];
    delete[] mpBinPaths[BIN];
    return bRet? EXIT_SUCCESS : EXIT_FAILURE;
}

/**
  * This is the main loop of LibFilter. It iterates on the directories
  * of the core repo.
  * @return  Signals failure (false) or success.
  */
bool LibFilter::iterateOnDirs()
{
    if (setenv("SD_LIBRARY_PATH", mpBinPaths[LIB], 0))
    {
        printf("Failed to set SD_LIBRARY_PATH\n");
        return false;
    }

    DIR* pLiboRoot = opendir(msLiboPath);
    if (pLiboRoot == NULL)
    {
        printf("Failed to open LibO's root dir (%s)\n", msLiboPath);
        return false;
    }

    for (dirent* pDir; pDir = readdir(pLiboRoot);)
    {
        if (pDir->d_type != DT_DIR)
            continue;

        char* sDirPath = new char[mnLiboPathLen + strlen(pDir->d_name) + 1];
        strcpy(sDirPath, msLiboPath);
        strcat(sDirPath, pDir->d_name);

        DIR* pOpenedDir = opendir(sDirPath);
        if (pOpenedDir == NULL)
        {
	        printf("Failed to open directory %s\n", sDirPath);
	        delete[] sDirPath;
	        continue;
        }

        for (dirent* pDirEntry; pDirEntry = readdir(pOpenedDir);)
            processDirEntry(pDirEntry, sDirPath);

        closedir(pOpenedDir);
        delete[] sDirPath;
    }
    closedir(pLiboRoot);
    return true;
}

/**
  * Processes the entries in the directories of the core repo. Searches for
  * GnuMakefiles that are responsible for building executables or libraries.
  * @param   pDirEntry       The directory entry (hopefully a proper makefile).
  * @param   sParentDirName  The name of the directory that contains the entry.
  *                          It is required to build the path of the makefile.
  */
void LibFilter::processDirEntry(dirent* pDirEntry, char* sParentDirName)
{
    char* sBinName;
    if (NULL == (sBinName = getBinNameIfProperMakefile(pDirEntry)))
        return;

    vector<char*> vUnusedLibs;
    if (!getUnusedLibraries(sBinName, vUnusedLibs))
    {
        delete[] sBinName;
        return; // go on with the next binary in case of failure
    }

    char* sFullMkName = new char[strlen(sParentDirName) + strlen(pDirEntry->d_name) + 2];
    strcpy(sFullMkName, sParentDirName);
    strcat(sFullMkName, "/");
    strcat(sFullMkName, pDirEntry->d_name);

    updateMakefile(sFullMkName, vUnusedLibs);

    delete[] sBinName;
    delete[] sFullMkName;
}

/**
  * Checks if the given directory entry is a GnuMakefile responsible for
  * building a binary. If the name of the entry matches the pattern
  * `{Library, Executable}_<name>.mk', then it fetches the name of the binary and
  * returns the path of it in solver.
  * @param   pDirEntry   The directory entry (hopefully a proper makefile).
  * @return  The path of the binary or NULL, if pDirEntry is not a proper makefile.
  */
char* LibFilter::getBinNameIfProperMakefile(dirent* pDirEntry)
{
    char* sEntryName = pDirEntry->d_name;
    size_t nEntryNameLen = strlen(sEntryName);
    LibOrBin eType;

    if (strcmp(&sEntryName[nEntryNameLen - 3], ".mk") != 0)
        return NULL;

    if (strncmp(sEntryName, "Library_", 8) == 0)
        eType = LIB;
    else if (strncmp(sEntryName, "Executable_", 11) == 0)
        eType = BIN;
    else
        return NULL;

    size_t nBinNameOffset = eType == LIB? 8 : 11;
    size_t nShortLen = nEntryNameLen - nBinNameOffset - 3;
    char* sBinName = new char[mnBinPathLen + nShortLen + (eType == LIB? 9 : 1)];
    strcpy(sBinName, mpBinPaths[eType]);

    if (eType == LIB)
    {
        strcat(sBinName, "lib");
        strncpy(&sBinName[mnBinPathLen + 3], &sEntryName[nBinNameOffset], nShortLen);
        sBinName[mnBinPathLen + 3 + nShortLen] = '\0';
        strcat(sBinName, "lo.so");
    }
    else
    {
        strncpy(&sBinName[mnBinPathLen], &sEntryName[nBinNameOffset], nShortLen);
        sBinName[mnBinPathLen + nShortLen] = '\0';
    }
    return sBinName;
}

/**
  * Given the path of a binary this function gets the "unused direct dependencies"
  * of it with the Gnu Binutil `ldd'. The unnecessary libraries' short names goes
  * to the second parameter.
  * @param   sBinName    The path of the binary to process.
  * @param   rLibraries  A reference to a vector which will contain
  *                      the short names of the unnecessary libraries.
  * @return  Signals failure (false) or success.
  */
bool LibFilter::getUnusedLibraries(char* sBinName, vector<char*>& rLibraries)
{
    char* sCommand = new char[strlen(sBinName) + 20];
    strcpy(sCommand, "ldd -u ");
    strcat(sCommand, sBinName);
    strcat(sCommand, " 2>/dev/null");

    FILE* pPipe = popen(sCommand, "r");
    delete[] sCommand;
    if (pPipe == NULL)
    {
        printf("Opening pipe for 'ldd -u %s' failed.\n", sBinName);
        return false;
    }

    char sBuffer[MAXLEN];
    fgets(sBuffer, MAXLEN, pPipe);
    if (strcmp(sBuffer, "Unused direct dependencies:\n") != 0)
    {
        pclose(pPipe);
        return false;
    }

    // Ignore the second line
    fgets(sBuffer, MAXLEN, pPipe);

    while (fgets(sBuffer, MAXLEN, pPipe) != NULL)
    {
        // Ignore libs not presented in mpBinPaths[LIB]
        if (strncmp(mpBinPaths[LIB], &sBuffer[1], mnBinPathLen) != 0)
            continue;

        // Fetch the short name of the library symbol
        size_t nLibNameLen = strlen(sBuffer) - mnBinPathLen - 2; //'\t'[…]'\n'
        char* sLibName = new char[nLibNameLen + 1];
        strncpy(sLibName, &sBuffer[mnBinPathLen + 1], nLibNameLen);

        if (!isLibNameProper(sLibName, nLibNameLen))
        {
            delete[] sLibName;
            continue;
        }

        memmove(sLibName, &sLibName[3], nLibNameLen - 8);
        sLibName[nLibNameLen - 8] = '\0';
        rLibraries.push_back(sLibName);
    }
    pclose(pPipe);
    return true;
}

/**
  * Deletes the entries of the unnecessary library dependencies in the makefile
  * of the currently processed binary.
  * @param  sMkName     The path of the makefile to update.
  * @param  rUnusedLibs The short names of the unnecessary library dependencies.
  */
void LibFilter::updateMakefile(char* sMkName, vector<char*>& rUnusedLibs)
{
    char* sTempMkName = new char[strlen(sMkName) + 2];
    strcpy(sTempMkName, sMkName);
    strcat(sTempMkName, "~");

    FILE* pTempMk = fopen(sTempMkName, "w");
    if (pTempMk == NULL)
    {
        printf("Failed to open temporary file %s\n", sTempMkName);
        delete[] sTempMkName;
        return;
    }

    FILE* pOrigMk = fopen(sMkName, "r");
    if (pOrigMk == NULL)
    {
        printf("Failed to open %s\n", sMkName);
        delete[] sTempMkName;
        return;
    }

    // Making patterns from the short names (<name> + ' ' + '\' + '\n')
    for (vector<char*>::iterator it = rUnusedLibs.begin(); it != rUnusedLibs.end(); ++it)
        strcat(*it, " \\\n"); // there are 7 empty places behind the name in the buffer

    char sBuffer[MAXLEN];
    while (fgets(sBuffer, MAXLEN, pOrigMk) != NULL)
    {
        vector<char*>::iterator it = rUnusedLibs.begin();
        bool bMatchFound = false;
        while (!bMatchFound && it != rUnusedLibs.end())
        {
            if (strlen(*it) <= strlen(sBuffer))
                bMatchFound = 0 == strcmp(*it, &sBuffer[strlen(sBuffer) - strlen(*it)]);
            ++it;
        }
        if (!bMatchFound)
            fputs(sBuffer, pTempMk);
    }

    fclose(pTempMk);
    fclose(pOrigMk);
    rename(sTempMkName, sMkName);

    for (vector<char*>::iterator it = rUnusedLibs.begin(); it != rUnusedLibs.end(); ++it)
        delete *it;
    delete sTempMkName;
}

/**
  * Checks if the given library name matches the pattern `lib*lo.so'.
  * @param  sLibName    The name of the library to check.
  * @param  nLen        The length of sLibName.
  * @return True, if sLibName is proper (matches the pattern above).
  */
inline bool LibFilter::isLibNameProper(char* sLibName, size_t nLen)
{
    return strncmp(sLibName, "lib", 3) == 0 &&
           strncmp(&sLibName[nLen - 5], "lo.so", 5) == 0;
}
